/**
 (c) James Gregson (james.gregson@gmail.com), 2013
 Free for academic and personal use with proper attribution, do not redistribute,
 use at your own risk.
 
 Sample (and simple) implementation of Stochastic Deconvolution, from the
 CVPR 2013 poster of the same name.  This is NOT the code used in the paper,
 it is only intended to illustrate the algorithm and provide other
 researchers/hobbyists with a starting point for experimenting with the method.
 
 The sample implementation DOES NOT implement everything from the paper,
 e.g. the following are not handled:
 - Saturation
 - Boundary conditions
 - Color images
 - Clamping to positive values
 - Most of the priors from the paper
 
 Requires ImageMagick for image IO via CImg.  Compiled and tested under OS-X
 with XCode, Windows users will need to provide a definition of the drand48()
 function.  Build (unix/os-x) by executing the following commands from the
 build/ subdirectory:
 
 $> cmake ../code
 $> make
 
 The binary will be placed in the bin/ subdirectory along with the input
 file dandelion.jpg. Run the code with the following command:
 
 $> ./stochastic_deconvolution_sample
 
 The executable will generate a number of image file: i) blurred.png, the
 synthetically blurred grayscale version of dandelion.jpg, ii) intrinsic.png,
 the debarred result and iii) ground_truth.png, the unblurred grayscale version
 of dandelion.jpg.
 
 The code is extensively commented in an effect to make it straightforward to
 understand what is being done and why. I recommend stepping through with a
 debugger to get a feeling for how it works.
 
 The algorithm deblurs images using the following model:
 
 p* = argmin_p F(p) = || A p - q ||_2^2 + gamma Gamma(p)
 
 where A is a matrix represeting a blur operation, p & q are respectively the
 intrinsic (solution) image and measured blurry image, gamma is a weight and
 Gamma(p) is a regularization function.
 
 The algorithm generates sample points at discrete locations in the blurry image
 and attempts to splat small quantities (e_d) of 'brightness' so as to minimize the
 objective function F(p). Small support of the blur operation A and regularizer
 function Gamma(p) make the effects of these local operations on F(p) relatively
 inexpensive.  The key point of the algorithm is to use a Metropolis-Hastings
 like sampling heuristic to generate sample locations that are likely to make
 fast progress. Reasonably fast convergence is achieved by scheduling the
 magnitude of e_d, although this can have theoretical implications on provable
 convergence for even smooth objectives, in practice it appears to work quite well.
 
 Overall the method is a heuristically-driven stochastic coordinate
 descent.  The same convergence theory applies here as to coordinate-descent,
 provided all pixels are checked before reducing e_d.  Convergence on non-smooth
 objectives is not guaranteed, but again, works reasonably well in practice.
 
 More information is available in the paper. Please contact me if you have
 any questions/comments.
 
 James Gregson
*/
#include<cmath>
#include<cstdlib>
#include<iostream>

// I use the excellent CImg library (http://cimg.sourceforge.net/) for image IO.
#include<CImg.h>
using namespace cimg_library;

#include"stochastic_deconvolution.h"

///////////////////////////////////////////////////////////////////////////////
// Method Parameters //////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
const double reg_weight     = 0.0005;   // regularizer weight
const double sigma          = 4.0;      // mutation standard deviation
const double reset_prob     = 0.005f;   // russian roulette chain reset probability
const int    num_iterations = 100;      // number of 'iterations', mostly for output
double ed                   = 0.025;     // starting deposition energy

///////////////////////////////////////////////////////////////////////////////
// Regularizer and PSF setup //////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// define a very simple PSF with 9 equal entries simulating motion blur
const int    psf_cnt = 9;
const double psf_v[] = { 1.0/9.0, 1.0/9.0, 1.0/9.0, 1.0/9.0, 1.0/9.0, 1.0/9.0, 1.0/9.0, 1.0/9.0, 1.0/9.0 };
const int    psf_x[] = { -4, -3, -2, -1, 0, 1, 2, 3, 4 };
const int    psf_y[] = { -4, -3, -2, -1, 0, 1, 2, 3, 4 };

// Define a stencil for evaluating the regularizer energy change. this is the
// only 'tricky' part of the algorithm: in order to evaluate the change in
// regularizer energy, it is necessary to evaluate the regularizer for every
// pixel whose stencil is affected by adding energy to a given (anchor) pixel.
// For this demo, the isotropic TV regularizer is used. Image gradients are
// computed with backward differences, so the anchor pixel and the top and
// right pixels will be affected by adding energy at the anchor. Basically
// whatever pixels are used by the regularizer should have their coordinates
// reflected about the central pixel to avoid introducing error.
// E.g., A=anchor, L=left, B=bottom, T=top, R=right:
//
//                                   T
//                                   |
//                                   |
//           L --- A       ==>       A -- R
//                 |
//                 |
//                 B
//
//        pixels accessed         pixels affected
//         by regularizer         by splatting at A
const int    reg_cnt = 3;
const int    reg_x[] = {  0,  1,  0 };
const int    reg_y[] = {  0,  0,  1 };

///////////////////////////////////////////////////////////////////////////////
// Forward declarations of callbacks and auxillary functions //////////////////
///////////////////////////////////////////////////////////////////////////////
// callbacks used by the sampling routine
void splat( sd_data *data, sd_sample *x, double weight );
void copy( sd_data *data, sd_sample *x, sd_sample *y );
double evaluate( sd_data *data, sd_sample *x );
void mutate( sd_data *data, sd_sample *x, sd_sample *y );

// auxilliary routines used by the callbacks
CImg<double> load_grayscale( const char *filename );
CImg<double> blur_image( CImg<double> &img );
void sample_normal( double &X, double &Y );
bool inside_image( sd_data *data, int x, int y );
double data_energy( sd_data *data, int x, int y );
double regularizer_energy( sd_data *data, int x, int y );


///////////////////////////////////////////////////////////////////////////////
// Entry point ////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
int main( int argc, char **argv ){
    // uncomment to allow debugging in XCode
    cimg::imagemagick_path( "/usr/local/bin/convert" );
    
    // load the input image, blur it to get a synthetic input and
    // set the initial guess for the solution equal to the synthetically
    // blurred image
    CImg<double> input = load_grayscale( "dandelion.jpg" );
    CImg<double> blurred = blur_image( input );
    CImg<double> solution = blurred;
    
    // initialize the data structure for the stochastic_deconvolution routine
    sd_data data;
    data.input = blurred;
    data.intrinsic = solution;
    data.blurred = blur_image( solution );
    
    // create the callback structure
    sd_callbacks cb;
    cb.copy = copy;
    cb.evaluate = evaluate;
    cb.mutate = mutate;
    cb.splat = splat;
    
    // actually perform the deconvolution, look for 100 iterations, applying
    // width*height mutations.  check the acceptance rate and if it drops below
    // 10%, scale the deposition energy down by a factor and continue
    double accept_rate = 0.0;
    for( int k=0; k<num_iterations; k++ ){
        accept_rate = stochastic_devonvolution( &data, &cb, ed, input.width()*input.height() );
        std::cout << "iteration " << k+1 << " of " << num_iterations << ", acceptance rate: " << accept_rate << ", ed: " << ed << std::endl;
        if( accept_rate < 0.4 )
            ed *= 0.5f;
    }
    
    // write out the grayscale ground-truth image, the synthetically
    // blurred image and the deconvolved intrinsic image
    (input.get_max(0.0)*255.0).save("ground_truth.png");
    (data.input.get_max(0.0)*255.0).save("blurred.png");
    (data.intrinsic.get_max(0.0).get_min(1.0)*255.0).save("intrinsic.png");
    
    // done!
    return 0;
}

// loads an image and converts it to grayscale by averaging
CImg<double> load_grayscale( const char *filename ){
    CImg<double> input(filename);
    CImg<double> gray = input.get_channel(0);
    for( int i=1; i<input.spectrum(); i++ ){
        gray += input.get_channel(i);
    }
    return gray.get_normalize( 0.0f, 1.0f );
}

// blurs the input image using the hardcoded PSF above, simply
// splats the input image into the blurred image according to
// the PSF
CImg<double> blur_image( CImg<double> &img ){
    int tx,ty;
    CImg<double> blurred( img.width(), img.height() );
    blurred.fill(0.0f);
    for( int i=0; i<img.width(); i++ ){
        for( int j=0; j<img.height(); j++ ){
            for( int k=0; k<psf_cnt; k++ ){
                tx=i+psf_x[k];
                ty=j+psf_y[k];
                if( tx >= 0 && ty >= 0 && tx < img.width() && ty < img.height() ){
                    blurred(tx,ty) += psf_v[k]*img(i,j);
                }
            }
        }
    }
    return blurred;
}

///////////////////////////////////////////////////////////////////////////////
// Stochastic Deconvolution Callbacks /////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// takes a sample and splats its energy into the intrinsic image at the
// sample location, also splats the psf*sample_energy into the blurred
// intrinsic image to allow evaluating the change in data-term energy
void splat( sd_data *data, sd_sample *x, double weight ){
    int tx, ty;
    data->intrinsic(x->x,x->y) += weight*x->ed;
    for( int i=0; i<psf_cnt; i++ ){
        tx = x->x + psf_x[i];
        ty = x->y + psf_y[i];
        if( tx > 0 && tx < data->blurred.width() && ty > 0 && ty < data->blurred.height() ){
            data->blurred(tx,ty) += weight*x->ed*psf_v[i];
        }
    }
}

// copies sample x into sample y
void copy( sd_data *data, sd_sample *x, sd_sample *y ){
    y->x = x->x;
    y->y = x->y;
    y->ed = x->ed;
}

// evaluates the change in the objective function due to applying the sample
// x with either positive or negative weighting. Simply checks the change in
// data+regularizer energy after splatting the sample with positive and
// negative weights
double evaluate( sd_data *data, sd_sample *x ){
    double init=0.0, plus=0.0, minus=0.0;
    double de_plus, de_minus;
        
    // start by computing the initial energy
    init = data_energy( data, x->x, x->y );
    for( int i=0; i<reg_cnt; i++ )
        init += regularizer_energy( data, x->x+reg_x[i], x->y+reg_y[i] );
    
    // splat the sample and evaluate the change
    // to the data fitting and regularizer energy
    splat( data, x, 1.0 );
    plus = data_energy( data, x->x, x->y );
    for( int i=0; i<reg_cnt; i++ )
        plus += regularizer_energy( data, x->x+reg_x[i], x->y+reg_y[i] );
    splat( data, x, -2.0 );  // remove twice the energy to get negative sample effect
    
    // compute the negative value
    minus = data_energy( data, x->x, x->y );
    for( int i=0; i<reg_cnt; i++ )
        minus += regularizer_energy( data, x->x+reg_x[i], x->y+reg_y[i] );
    splat( data, x, 1.0 ); // splat back the sample energy to remove all splatting effects
    
    // evalute the change in energy
    de_plus  = init - plus;
    de_minus = init - minus;
    
    // if the negative energy provides the best
    // improvement, negate ed and return the energy
    // change
    if( de_minus > de_plus ){
        x->ed = -x->ed;
        return de_minus;
    }
    
    // return the positive energy change
    return de_plus;
}

// mutates sample x by either i) offseting the current sample by a
// normally distributed offset or ii) completely reseting the chain.
// (ii) is done by Russian Roulette.  
void mutate( sd_data *data, sd_sample *x, sd_sample *y ){
    double dx, dy;
    y->ed = x->ed;
    if( drand48() >= reset_prob ){
        // normally distributed offset, keep sampling until a
        // new sample point is found that is not coincident with
        // the current point
        while( true ){
            sample_normal( dx, dy );
            y->x = int(double(x->x) + sigma*dx + 0.5);
            y->y = int(double(x->y) + sigma*dy + 0.5);
            if( (y->x != x->x || y->y != x->y) && inside_image( data, y->x, y->y ) )
                break;
        }
    } else {
        // reset the chain entirely
        y->x = drand48()*data->intrinsic.width();
        y->y = drand48()*data->intrinsic.height();
    }
}

// Computes the energy of the data term for a pixel at [x,y]. Given a matrix
// A representing convolution by the transpose of the PSF, the data term is:
//
// || A p - q ||_2^2
//
// where p is the intrinsic image and q is the measured (or for this demo,
// synthetic) blurry input image.  Stochastic deconvolution maintains a
// stored image b = A p (i.e. the current estimate of the intrinsic image
// blurred by the PSF, called data->blurred in the code below).  A change
// at [x,y] can only influence pixels in the support of the PSF, so only
// a small portion of the data term needs to be evaluated to see the effect
// of splatting a small change at [x,y]
double data_energy( sd_data *data, int x, int y ){
    int tx, ty;
    double delta, sum = 0.0;
    for( int i=0; i<psf_cnt; i++ ){
        tx = x + psf_x[i];
        ty = y + psf_y[i];
        if( inside_image( data, tx, ty ) ){
            delta = data->blurred(tx,ty) - data->input(tx,ty);
            sum += delta*delta;
        }
    }
    return sum;
}

// evaluates the regularizer function at [x,y]. It is VITAL that the
// pixels accessed by the regularizer function are included in the
// regularizer stencil at the top of this file, see the discussion
// there for details.
//
// this regularizer is simply isotropic TV
double regularizer_energy( sd_data *data, int x, int y ){
    double dx=0.0, dy=0.0;
    dx = x <= 0 ? 0.0 : data->intrinsic( x, y ) - data->intrinsic( x-1, y );
    dy = y <= 0 ? 0.0 : data->intrinsic( x, y ) - data->intrinsic( x, y-1 );
    return reg_weight*sqrt( dx*dx + dy*dy );
}

// draws two samples from a Gaussian (normal) distribution using
// a logarithmic transform. Note that it is important that the
// Gaussian is not truncated or ergodicity of the method
// disappears.
void sample_normal( double &X, double &Y ){
    double u=drand48(), v=drand48();
    double lnu = sqrt( -2.0*log(u) );
    X = lnu*cos(2.0*M_PI*v);
    Y = lnu*sin(2.0*M_PI*v);
}

// returns true if the [x,y] is inside the image
bool inside_image( sd_data *data, int x, int y ){
    return x >= 0 && x < data->blurred.width() && y >= 0 && y < data->blurred.height();
}
