#if !defined(STOCHASTIC_DECONVOLUTION_H)
#define STOCHASTIC_DECONVOLUTION_H

/*
 (c) James Gregson (james.gregson@gmail.com), 2013
 Free for academic and personal use with proper attribution, do not redistribute,
 use at your own risk.
 
 Sample (and simple) implementation of Stochastic Deconvolution, from the
 CVPR 2013 poster of the same name.  This is NOT the code used in the paper,
 it is only intended to illustrate the algorithm and provide other
 researchers/hobbyists with a starting point for experimenting with the method.
 
 Header file for stochastic deconvolution. Just defines a few data-types and 
 the callback function prototypes along with the stochastic_deconvolution
 function
*/

// I use the CImg library for IO and as an internal data-structure
#include<CImg.h>
using namespace cimg_library;

// defines a sample for use by the method. x and y are the pixel
// location, ed is the energy of the sample
typedef struct {
    int     x;
    int     y;
    double  ed;
} sd_sample;

// data structure to hold the input blurry image, the
// sharp intrinsic image (that gets computed) and the
// blurred version of the intrinsic image that is
// used to compute the change in data-term
typedef struct {
    CImg<double>     input;
    CImg<double>     blurred;
    CImg<double>     intrinsic; 
} sd_data;

// structure of callback functions
typedef struct {
    // Evaluates the change in energy due to a sample x being splatted with
    // either positive or negative weight. Returns the best energy change
    // and sets the sign of x->ed to match
    double (*evaluate)( sd_data *data, sd_sample *x );
    
    // generates a new sample y from an existing sample x by a mutation.
    // should be ergodic, i.e. all points in domain sampled with finite
    // probability at all times.
    void   (*mutate)( sd_data *data, sd_sample *x, sd_sample *y );
    
    // splats a sample into the intrinsic image and blurred-intrinsic
    // image to 'record' a sample that has improved the objective function
    void   (*splat)( sd_data *data, sd_sample *x, double weight );
    
    // copies sample x into sample y
    void   (*copy)( sd_data *data, sd_sample *x, sd_sample *y );
} sd_callbacks;

// the decovolution routine itself, deceptively simple when you
// look at the code in stochastic_deconvolution.cpp
double stochastic_devonvolution( sd_data *data, sd_callbacks *cb, double ed, int n_mutations );

#endif